"""
Copyright (c)2013 The MITRE Corporation

This software is the copyrighted work of MITRE.  No ownership or other proprietary interest in this software is granted to you other than what is granted in this license.     
 
MITRE IS PROVIDING THE SOFTWARE "AS IS" AND ACCORDINGLY MAKES NO WARRANTY, EXPRESS OR IMPLIED, AS TO THE ACCURACY, CAPABILITY, EFFICIENCY, MERCHANTABILITY, OR FUNCTIONING OF THE SOFTWARE AND DOCUMENTATION.  IN NO EVENT WILL MITRE BE LIABLE FOR ANY GENERAL, CONSEQUENTIAL, INDIRECT, INCIDENTAL, EXEMPLARY, OR SPECIAL DAMAGES RELATED TO THE SOFTWARE, EVEN IF MITRE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.             
 
You accept this software on the condition that you indemnify and hold harmless MITRE, its Board of Trustees, officers, agents, and employees, from any and all liability damages to third parties, including attorneys' fees, court costs, and other related costs and expenses, arising out of your use of this software irrespective of the cause of said liability.

Written by Sam Cornwell, scornwell@mitre.org

Usage:
  protections.py (per-version | per-file [-s]) [--pr-relaxed] [-pid] [-o FILE] <csv-directory> ...
  protections.py (-h | --help)

Description:
  Determine SMRAM and BIOS writability based on the CSV files generated by Copernicus.
  Attempts to parse all CSV files in the given CSV directories.
  
Options:
  per-version        aggregate output on a product name and bios vendor/version basis
  per-file           aggregate output on a input file basis
  --pr-relaxed       do not require that the PR registers cover the entire BIOS region
  -s                 include column for the source file
  -p                 include columns for the individual protections checked
  -i                 filter out machines with inconsistent protections across multiple CSV files
  -d                 deduplicate by hostname, processing only the newest file from any particular host
  -o, --output FILE  if given, will output the results into a CSV-formatted file  
"""

from docopt import docopt

import sys
import os
import glob
import ast
import string
from collections import defaultdict
from itertools import izip

COP_HDATA_SMRR_UNSUPPORTED = 12

class Version(object):
	def __init__(self, values):
		self.product_name = values.get('PRODUCT_NAME', '').strip()
		self.bios_vendor = values.get('BIOS_VENDOR', '').strip()
		self.bios_version = values.get('BIOS_VERSION', '').strip()
		
	def __cmp__(self, other):
		if self.bios_vendor != other.bios_vendor:
			return cmp(self.bios_vendor, other.bios_vendor)
		elif self.product_name != other.product_name:
			return cmp(self.product_name, other.product_name)
		else:
			return cmp(self.bios_version, other.bios_version)
	
	def __eq__(self, other):
		return cmp(self, other) == 0
	
	def __keys(self):
		return (self.bios_vendor, self.product_name, self.bios_version)
		
	def __hash__(self):
		return hash(self.__keys())
	
class ParsedData(object):
	pr_relaxed = False

	def __init__(self, values, filename = None):
		self.values = values
		self.filename = filename
		self.d_lck = d_lck_set(values)
		self.smrr_enabled = smrr_enabled(values)
		self.bios_pr = pr_protects_bios(values, ParsedData.pr_relaxed)
		self.bioswe_lock = bool_str_to_bool(values['BIOSWE_LOCK'])
		self.smm_bwp = smm_bwp_set(values)
	
	@property
	def smram_unlocked(self):
		return not self.smrr_enabled

	@property
	def bios_unlocked(self):
		return not self.bios_pr and not (not self.smram_unlocked and (self.bioswe_lock or self.smm_bwp))
		
	def equal_protections(self, other):
		return self.d_lck == other.d_lck and self.smrr_enabled == other.smrr_enabled and self.bios_pr == other.bios_pr and self.bioswe_lock == other.bioswe_lock and self.smm_bwp == other.smm_bwp
	
class AggregatedProtections(object):
	def __init__(self, machines = None):
		self.count = 0
		self.d_lck = 0
		self.smrr_enabled = 0
		self.bios_pr = 0
		self.bioswe_lock = 0
		self.smm_bwp = 0
		self.smram_unlocked = 0
		self.bios_unlocked = 0
		
		if machines:
			self.aggregate(machines)
	
	# all the variables casted as ints are originally bools, so they will be 1 (True) or 0 (False)
	def add_protections(self, data):
		self.count += 1
		self.d_lck += int(data.d_lck)
		self.smrr_enabled += int(data.smrr_enabled)
		self.bios_pr += int(data.bios_pr)
		self.bioswe_lock += int(data.bioswe_lock)
		self.smm_bwp += int(data.smm_bwp)
		self.smram_unlocked += int(data.smram_unlocked)
		self.bios_unlocked += int(data.bios_unlocked)
	
	def aggregate(self, machines):
		for machine in machines:
			self.add_protections(machine)

def csv_data_to_dict(data):
	d = {}
	fields = data.split(',')
	
	for field in fields:
		try:
			(key, value) = field.split('=', 1)
			d[key] = value
		except:
			# ignore this field I guess
			pass
	
	return d

def hex_str_to_long(s):
	#return reduce(lambda acc, (i, c): acc | (ord(c) << 8*i), enumerate(s.decode('hex')[::-1]), 0)
	return int(s, 16)

def bool_str_to_bool(s):
	return ast.literal_eval(s)
	
def get_bit_field(value, bottom, num_bits):
	if isinstance(value, (int, long)):
		int_value = value
	elif isinstance(value, str):
		int_value = hex_str_to_long(value)

	mask = ((1 << num_bits) - 1) << bottom
	return (int_value & mask) >> bottom

def bit_is_set(value, bit):
	return bool(get_bit_field(value, bit, 1))

def sanitize_region((base, limit)):
	if (base > limit):
		return None
	else:
		return base, limit

# works for both PR and FREG registers
def region_str_to_region(reg_str):
	reg_value = hex_str_to_long(reg_str)
	base = (reg_value & 0x00001fff) << 12
	limit = (((reg_value & 0x1fff0000) >> 4) | 0x00000fff)
	return sanitize_region((base, limit))

def region_intersect((base1, limit1), (base2, limit2)):
	sanitize_region((max(base1, base2), min(limit1, limit2)))

#this actually isn't necesary anymore, can use region_set_union instead
def region_merge((base1, limit1), (base2, limit2)):
	# this could probably be optimized, but it's easier to create the algorithm
	# by simply eliminating every case one at a time
	# the cases where one region is completely in the other
	if base1 >= base2 and limit1 <= limit2:
		return [(base2, limit2)]
	if base2 >= base1 and limit2 <= limit1:
		return [(base1, limit1)]
	# the cases where regions are completely separated
	if limit1 < base2 - 1 or limit2 < base1 - 1:
		return [(base1, limit1), (base2, limit2)]
	# finally the cases where the regions partially overlap
	if base1 < base2:
		return [(base1, limit2)]
	else:
		return [(base2, limit1)]

def region_cmp((base1, limit1), (base2, limit2)):
	return cmp(base1, base2) if base1 != base2 else cmp(limit1, limit2)

def region_set_union(regions):
	regions = sorted(regions, cmp=region_cmp)
	merged_regions = []
	while len(regions) >= 2:
		base1, limit1 = regions.pop(0)
		base2, limit2 = regions[0]
		
		if (base1 == base2):
			continue
		
		if (limit1 >= base2 - 1):
			regions[0] = (base1, max(limit1, limit2))
		else:
			merged_regions.append((base1, limit1))
		
	if (regions):
		merged_regions.append(regions.pop(0))
	
	return merged_regions

# region1 - region2
# warning: reading may cause aneurysm
def region_diff((base1, limit1), (base2, limit2)):
	if base1 < base2:
		if limit2 < limit1:
			return [(base1, base2 - 1), (limit2 + 1, limit1)]
		else:
			return [(base1, min(base2 - 1, limit1))]
	else:
		if limit2 >= limit1:
			return []
		else:
			return [(max(limit2 + 1, base1), limit1)]

def region_set_diff(region_set, region2):
	return region_set_union(sum([region_diff(region1, region2) for region1 in region_set], []))
	
def pr_protects_bios(values, relaxed=False):
	# these all have pr0 - pr2
	if values['CHIPSET_FAMILY'] in ['ICH7', 'ICH8', 'ICH9', 'ICH10', 'PCH']:
		pr_set = ['PR0', 'PR1', 'PR2']
		# all except ICH7 have pr3 and pr4 as well
		if values['CHIPSET_FAMILY'] != 'ICH7':
			pr_set += ['PR3', 'PR4']
	else:
		# we don't know what PRs are available, if any, so just return False
		return False

	# I put bios_range into a list to start because it will potentially get split up into a list of pieces
	# BIOS region is irrelevant if we are not in relaxed mode
	if not relaxed:
		if bool_str_to_bool(values['FLASH_DESCRIPTOR_MODE']):
			bios_regions = [region_str_to_region(values['FREG1'])]
		else:
			bios_regions = [(0, hex_str_to_long(values['FLASH_CHIP_SIZE']) - 1)]
		
	for pr in pr_set:
		# check if this PR offers write-protection
		if not bit_is_set(values[pr], 31):
			continue
		elif relaxed:
			# if the PR offers write-protection, then that's good enough from a relaxed perspective!
			return True
		
		# parse out the base, limit from the PR register
		pr_region = region_str_to_region(values[pr])
		# the region could be invalid (unused)
		if not pr_region:
			continue
		
		# subtract off the protected region from the bios region parts
		# which have not been covered yet by another PR
		bios_regions = region_set_diff(bios_regions, pr_region)

	if relaxed:
		return False
	else:
		# the bios region set should be empty if it has been completely covered by PRs
		return not bios_regions

def smm_bwp_set(values):
	# this bit was first seen in the Intel 5 series, which is also the first version with a PCH
	return values['CHIPSET_FAMILY'] == 'PCH' and bit_is_set(values['BIOS_CNTL'], 5)
	
def d_lck_set(values):
	return bit_is_set(values['SMRAMC'], 4)

def smrr_supported(values):
	return not bit_is_set(values['COP_RETURN_STATUS'], COP_HDATA_SMRR_UNSUPPORTED)
	
def smrr_enabled(values):
	return smrr_supported(values) and bit_is_set(values['SMRR_PHYSMASK'], 11)

def smm_mechanisms(values):
	lock_mechanisms = []
	
	#if d_lck_set(values) and smrr_enabled(values):
	if smrr_enabled(values):
		lock_mechanisms.append("D_LCK is set and SMRR interface is enabled.")

	return lock_mechanisms
	
def bios_mechanisms(values, smm_locked):
	lock_mechanisms = []

	# debating whether mechanism should also require smm to be locked
	if bool_str_to_bool(values['BIOSWE_LOCK']):
		lock_mechanisms.append("BIOS_WE bit is locked down by SMM.")

	# debating whether mechanism should also require smm to be locked
	if smm_bwp_set(values):
		lock_mechanisms.append("SMM_BWP bit set, BIOS locked down by SMM.")
	
	if pr_protects_bios(values):
		lock_mechanisms.append("Entire BIOS region is write-projected by one or more PR registers.")

	return lock_mechanisms

def print_vulnerability(checking, mechanisms):
	if mechanisms:
		print checking, "NOT Vulnerable"
		for mechanism in mechanisms:
			print "\t" + mechanism
	else:
		print checking, "Vulnerable"
		print "\t" + "No known write-locking mechanisms are properly enabled."

def parse_all_csvs(directories, verbose = False):
	csv_directories = [dir + os.sep if dir[-1] != os.sep else dir for dir in directories]
	filenames = sum([glob.glob(dir + '*.csv') for dir in csv_directories], [])
	
	parsed_data = []
	
	for filename in filenames:
		try:
			with open(filename, 'r') as f:
				values = csv_data_to_dict(f.read())
			parsed_data.append(ParsedData(values, filename))
		except Exception, e:
			# completely ignore any files we couldn't parse
			if verbose:
				print "Failed to parse", filename
	
	return parsed_data

def group_by_version(parsed_data):
	groups = defaultdict(list)
	
	for p in parsed_data:
		groups[Version(p.values)].append(p)
	
	return groups

def group_by_hostname(parsed_data):
	groups = defaultdict(list)
	
	for p in parsed_data:
		groups[p.values.get('HOSTNAME', '')].append(p)
	
	return groups

def timestamp_from_parsed_data(parsed_data):
	return int(parsed_data.filename.rsplit('.',1)[0].rsplit('-', 1)[1], 16)

def filter_inconsistent_machines(parsed_data):
	groups = group_by_hostname(parsed_data)
	if '' in groups:
		del groups['']
	
	for hostname, group in groups.iteritems():
		# I want to check here that the protections are all the same, so I compare them all to the first in the list
		if [p for p in group if not group[0].equal_protections(p)] != []:
			parsed_data = [p for p in parsed_data if p not in group]
	
	return parsed_data

def deduplicate_by_hostname(parsed_data):
	groups = group_by_hostname(parsed_data)
	no_hostname = groups.pop('', [])
	return [ sorted(group, key = timestamp_from_parsed_data, reverse = True)[0] for group in groups.itervalues()] + no_hostname

def get_max_len(arr):
	return max(map(len, arr))

def pad_to_max_len(s, arr):
	return s.ljust(get_max_len(arr), ' ')

# MAIN
if __name__ == '__main__':
	arguments = docopt(__doc__)
	
	if arguments['--pr-relaxed']:
		ParsedData.pr_relaxed = True

	parsed_data = parse_all_csvs(arguments['<csv-directory>'])

	# not sure if this should be before or after the summary calculations shortly below
	# I put it here on the off chance we remove all of the data, so that the next check will catch it
	if arguments['-i']:
		parsed_data = filter_inconsistent_machines(parsed_data)

	if not parsed_data:
		print "No valid CSV files found."
		sys.exit(0)

	if arguments['-d']:
		parsed_data = deduplicate_by_hostname(parsed_data)
	
	num_processed = len(parsed_data)
	num_smram_unlocked = sum((int(p.smram_unlocked) for p in parsed_data))
	num_bios_unlocked = sum((int(p.bios_unlocked) for p in parsed_data))

	'''
	if arguments['per-file']:
		for filename, (values, smm_mechanisms, bios_mechanisms) in parsed_data.iteritems():
			print "File:", filename
			print "Hostname:", values.get('HOSTNAME', '')
			print_vulnerability("SMRAM", smm_mechanisms)
			print_vulnerability("BIOS", bios_mechanisms)
			print #newline
	'''
	
	# set up base column names
	col_names = [
		pad_to_max_len("BIOS_VENDOR", (Version(p.values).bios_vendor for p in parsed_data)),
		pad_to_max_len("PRODUCT_NAME", (Version(p.values).product_name for p in parsed_data)),
		pad_to_max_len("BIOS_VERSION", (Version(p.values).bios_version for p in parsed_data)),
		"SMRAM_UNLOCKED",
		"BIOS_UNLOCKED"]
	
	# add column names specific to each command type
	if arguments['per-version']:
		col_names.insert(0, "COUNT")
	elif arguments['per-file']:
		col_names.insert(0, pad_to_max_len("HOSTNAME", (p.values.get('HOSTNAME', '') for p in parsed_data)))
	
	# these can be added for either command
	if arguments['-p']:
		col_names.extend(["D_LCK", "SMRR_ENABLED", "BIOS_PR", "BIOSWE_LOCK", "SMM_BWP"])
		
	# this is command-specific, but I want it to be the last column, so I add it separately here
	if arguments['per-file'] and arguments['-s']:
		col_names.append(pad_to_max_len("SOURCE_FILE", (p.filename for p in parsed_data)))
	
	try:
		outfile = open(arguments['--output'], 'w')
		outfile.write(','.join(map(string.strip, col_names)))
		outfile.write('\n')
	except IOError:
		print "Could not create output file", arguments['--output']
		outfile = None
	except Exception, e:
		# there was no output file passed
		outfile = None
	
	# table header
	print ' '.join(col_names)
	
	grouped_data = group_by_version(parsed_data)
	
	for version in sorted(grouped_data.keys(), key = lambda v: (v.bios_vendor.lower(), v.product_name.lower(), v.bios_version.lower())):
		# add in the basic fields
		if arguments['per-version']:
			agg = AggregatedProtections(grouped_data[version])
			group = [agg] # needed in case -p is set
			rows = [[str(agg.count),
				version.bios_vendor,
				version.product_name,
				version.bios_version,
				str(agg.smram_unlocked),
				str(agg.bios_unlocked)]]
		elif arguments['per-file']:
			group = sorted(grouped_data[version], key = lambda p: p.values.get('HOSTNAME', '').lower()) 
			rows = [[p.values.get('HOSTNAME', ''),
				version.bios_vendor,
				version.product_name,
				version.bios_version,
				str(p.smram_unlocked),
				str(p.bios_unlocked)] for p in group]

		# add in the individual protections if the command line argument was given
		if arguments['-p']:
			for row, p in izip(rows, group):
				row.extend([str(p.d_lck), str(p.smrr_enabled), str(p.bios_pr), str(p.bioswe_lock), str(p.smm_bwp)])
		
		# add in the source file if the command line argument was given
		if arguments['per-file'] and arguments['-s']:
			for row, p in izip(rows, group):
				row.append(p.filename)
		
		for row in rows:
			if outfile:
				outfile.write(','.join(row))
				outfile.write('\n')
			
			# pad it to make it look purdy
			row = map(lambda (col_val, col_name): col_val.ljust(len(col_name), ' '), zip(row, col_names))	
			print ' '.join(row)
	
	print #newline
	print num_processed, "CSV files successfully processed"
	print num_smram_unlocked, '(' + str(float(num_smram_unlocked)/float(num_processed) * 100) + '%)', "CSV files showing unlocked SMRAM"
	print num_bios_unlocked, '(' + str(float(num_bios_unlocked)/float(num_processed) * 100) + '%)', "CSV files showing unlocked BIOS"
